/*******************************************************************************
**                        backupprofile.cpp
**                             part of
**                konserve -- A small backup application
**			 -------------------
**  copyright: (C) 2002 - 2004 by Florian Simnacher
**  email    : simnacher AT gmx DOT de
*******************************************************************************/

/*******************************************************************************
**  This program is free software; you can redistribute it and/or             **
**  modify it under the terms of the GNU General Public License               **
**  as published by the Free Software Foundation; either version 2            **
**  of the License, or (at your option) any later version.                    **
**                                                                            **
**  This program is distributed in the hope that it will be useful,           **
**  but WITHOUT ANY WARRANTY; without even the implied warranty of            **
**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the             **
**  GNU General Public License for more details.                              **
**                                                                            **
**  You shougld have received a copy of the GNU General Public License        **
**  along with this program; if not, write to the Free Software               **
**  Foundation, Inc., 59 Temple Place - Suite 330, Boston,                    **
**  MA  02111-1307, USA.                                                      **
*******************************************************************************/

// Qt includes
#include <qfile.h>
#include <qfileinfo.h>
#include <qstringlist.h>
#include <qdir.h>

// KDE includes
#include <ktempfile.h>
#include <kmessagebox.h>
#include <klocale.h>
#include <kapplication.h>

// My includes
#include "backupprofile.h"
#include "helper.h"
#include <kdebug.h>

/*******************************************************************************
void __debug_print_kurl( KURL url, QString obj )
{
    kdDebug() << obj << "." << "directory(false, false): "
              << url.directory(false, false)
              << endl;
    kdDebug() << obj << "." << "fileName(false): " << url.fileName(false)
              << endl;
    kdDebug() << obj << "." << "path(): " << url.path() << endl;
    kdDebug() << obj << "." << "url(): " << url.url() << endl << endl;
}
*******************************************************************************/

////////////////////////////////////////////////////////////////////////////////
///////                  class Backup                                    ///////
////////////////////////////////////////////////////////////////////////////////

Backup::Backup( KURL surl,
                KURL aurl,
                bool readonly )
    : Dumpable( "Backup" /* , QString().setNum( (long) this, 16 ) */ ),
      mSourceUrl( surl ),
      mTmpSourceUrl(),
      mArchiveUrl( aurl ),
      mTmpArchiveUrl(),
      mbReadOnly( readonly ),
      mpArchive( 0L )
{
    kdDebug(0) << "Backup::Backup" << endl;
    // NOOP
}

Backup::~Backup()
{
    kdDebug(0) << "Backup::~Backup" << endl;
    // NOOP
}

void Backup::doBackup()
{
    kdDebug(0) << "Backup::doBackup" << endl;
    Q_ASSERT( mSourceUrl.isValid() );
    Q_ASSERT( mArchiveUrl.isValid() );
    Q_ASSERT( ! mbReadOnly );

    KTempFile tmpFile;
    mTmpSourceUrl.setPath( tmpFile.name() );
    tmpFile.unlink();

    if ( mSourceUrl.isLocalFile() )
    {
        mTmpSourceUrl = mSourceUrl;
        slotStartListing( 0L );
    }
    else
    {
        KIO::Job* job = KIO::copy( mSourceUrl, mTmpSourceUrl );
        connect(
            dynamic_cast<const QObject*>(job), SIGNAL( result( KIO::Job * ) ),
            SLOT( slotStartListing( KIO::Job *) )
            );
    }
}

QString Backup::dumpInternal()
{
    kdDebug(0) << "Backup::dumpInternal" << endl;
    QString tmp;

    tmp += QString( "mSourceUrl:  " ) + mSourceUrl.url()  + "\n";
    tmp += QString( "mArchiveUrl: " ) + mArchiveUrl.url() + "\n";
    if ( mbReadOnly )
        tmp += QString( "readonly" );
    else
        tmp += QString( "read/write" );

    return tmp;
}

void Backup::slotStartListing( KIO::Job* aJob)
{
    kdDebug(0) << "Backup::slotStartListing" << " aJob=" << aJob << endl;
    Q_ASSERT( 0L == mpArchive );
    Q_ASSERT( mTmpSourceUrl.isValid() );
    Q_ASSERT( mTmpSourceUrl.isLocalFile() );

    if ( 0L != aJob )
    {
        if ( aJob->error() )
        {
            aJob->showErrorDialog();
            return;
        }
    }
    KTempFile tmpFile;
    mTmpArchiveUrl.setPath( tmpFile.name() );
    tmpFile.unlink();

    mpArchive = new KTar( mTmpArchiveUrl.path(), "application/x-gzip" );
    mpArchive->open( IO_WriteOnly );

    if ( QFileInfo( mTmpSourceUrl.path() ).isFile() ||
         QFileInfo( mTmpSourceUrl.path() ).isSymLink())
    {
        KFileItem f( const_cast<const KURL&>( mTmpSourceUrl ),
                     KMimeType::defaultMimeType(),
                     KFileItem::Unknown );
        this->writeToArchive( f );
        mpArchive->close();
        delete mpArchive;
        mpArchive = 0L;
        copyArchive();
    }
    else if ( QFileInfo( mTmpSourceUrl.path() ).isDir() )
    {
        KIO::Job* job = KIO::listRecursive( mTmpSourceUrl, false, true );
        connect(
            job, SIGNAL( entries( KIO::Job*, const KIO::UDSEntryList&)),
            SLOT( slotGetEntriesAndPackEm( KIO::Job*, const KIO::UDSEntryList&))
            );
        connect(
            job, SIGNAL( result( KIO::Job* ) ),
            SLOT( slotFinishedPacking( KIO::Job* ) )
            );
    }
    else
    {
        Q_ASSERT( false );
    }

    Q_ASSERT( mTmpArchiveUrl.isValid() );
    Q_ASSERT( mTmpArchiveUrl.isLocalFile() );
}

void Backup::slotGetEntriesAndPackEm( KIO::Job*, const KIO::UDSEntryList& list )
{
    kdDebug(0) << "Backup::slotGetEntriesAndPackEm" << endl;
    Q_ASSERT( mTmpSourceUrl.isValid() );
    Q_ASSERT( mTmpSourceUrl.isLocalFile() );

    KIO::UDSEntryListConstIterator it=list.begin();
    for (; it != list.end(); ++it)
    {
        KFileItem f( *it, mTmpSourceUrl );

        if ( f.text() == "." || f.text() == ".." )
            continue;
        if ( f.isDir() )
            continue;
        else if ( f.isFile() )
            writeToArchive( f, true );
        else
            Q_ASSERT( false );
    }
}

void Backup::slotFinishedPacking( KIO::Job* aJob )
{
    kdDebug(0) << "Backup::slotFinishedPacking" << " aJob=" << aJob << endl;
    Q_ASSERT( QFileInfo( mTmpArchiveUrl.path() ).exists() );
    if ( aJob->error() )
    {
        aJob->showErrorDialog();
        return;
    }
    mpArchive->close();
    // who has invented the protected base class Dtor???
    delete dynamic_cast<KTar*>(mpArchive);
    mpArchive = 0L;
    copyArchive();
}

void Backup::copyArchive()
{
    kdDebug(0) << "Backup::copyArchive" << endl;
    Q_ASSERT( mArchiveUrl.isValid() );
    Q_ASSERT( QFileInfo( mTmpArchiveUrl.path() ).exists() );
    Q_ASSERT( mTmpArchiveUrl.isValid() );
    Q_ASSERT( mTmpArchiveUrl.isLocalFile() );

    KURL destUrl;
    if ( mArchiveUrl.fileName(false) == QString::null )
        destUrl = dotCheck( defaultArchiveName( mSourceUrl, mArchiveUrl ) );
    else
        destUrl = dotCheck( mArchiveUrl );

    KIO::Job* job = KIO::file_move( mTmpArchiveUrl, destUrl, -1, true );
    connect(
        job, SIGNAL( result( KIO::Job * ) ),
        SLOT( slotFinished( KIO::Job *) )
        );
}

void Backup::slotFinished( KIO::Job* aJob )
{
    kdDebug(0) << "Backup::slotFinished" << " aJob=" << aJob << endl;
    if ( aJob->error() )
    {
        aJob->showErrorDialog();
        emit finishedBackup(false);
        return;
    }
    emit finishedBackup(true);

    Q_ASSERT( !QFileInfo( mTmpArchiveUrl.path() ).exists() );
}

KURL  Backup::defaultArchiveName( KURL source, KURL backup )
{
    kdDebug(0) << "Backup::defaultArchiveName" << " source=" << source.url() << " backup=" << backup.url() << endl;
    QDateTime qdt = QDateTime::currentDateTime();

    return KURL(
        backup.url(1) + source.fileName() + "-" +
        qdt.toString( "yyyyMMddhhmmss" ) + ".tar.gz"
        );
}

KURL Backup::dotCheck( KURL inUrl )
{
    kdDebug(0) << "Backup::dotCheck" << " inUrl=" << inUrl.url() << endl;
    QString url_string = inUrl.url();
    int index = url_string.findRev( "/" );

    if ( -1 == index )
    {
        if ( url_string.startsWith( "." ) )
            return KURL( "dot" + url_string );
        else
            return inUrl;
    }
    if ( -1 < index )
    {
        QString rightstring = url_string.right( url_string.length() - index - 1 );
        QString leftstring  = url_string.left( index + 1 );

        if ( rightstring.startsWith( "." ) )
            return KURL( leftstring + "dot" + rightstring );
    }
    return inUrl;
}

void Backup::writeToArchive( KFileItem f, bool usetext )
{
    kdDebug(0) << "Backup::writeToArchive" << " usetext=" << (usetext ? "true" : "false" ) << endl;
    if ( f.isLink() )
    {
        QFile *file;

        if (usetext)
            file = new QFile( f.url().path() + f.name() );
        else
            file = new QFile( f.url().path() );

        Q_CHECK_PTR( file );
        Q_ASSERT( file->exists() );

        char  *c = new char[file->size()];

        if ( !file->open( IO_ReadOnly ) )
            return;
        file->readBlock( c, file->size() );

        if (usetext)
            mpArchive->writeFile(
                mSourceUrl.path() + f.text(), Helper::userName(),
                Helper::userGroup(), (unsigned int) file->size(), c
                );
        else
            mpArchive->writeFile(
                mSourceUrl.path(), Helper::userName(),
                Helper::userGroup(), (unsigned int) file->size(), c
                );
        file->close();
        delete[] c;
        delete file;
    }
    else if ( f.isFile() )
    {
        QFile *file;
        if (usetext)
            file = new QFile( f.url().path() + f.text() );
        else
            file = new QFile( f.url().path() );

        Q_CHECK_PTR( file );
        Q_ASSERT( file->exists() );
        char  *c = new char[file->size()];

        if ( !file->open( IO_ReadOnly ) )
            return;
        file->readBlock( c, file->size() );

        if (usetext)
            mpArchive->writeFile(
                mSourceUrl.path() + f.text(), Helper::userName(),
                Helper::userGroup(), (unsigned int) file->size(), c
                );
        else
            mpArchive->writeFile(
                mSourceUrl.path(), Helper::userName(),
                Helper::userGroup(), (unsigned int) file->size(), c
                );
        file->close();
        delete[] c;
        delete file;
    }
    else
    {
        Q_ASSERT( false ); // Should not happen
    }
}

////////////////////////////////////////////////////////////////////////////////
///////                  class BackupProfile                             ///////
////////////////////////////////////////////////////////////////////////////////

BackupProfile::BackupProfile( Backup*      backup,
                              QString      identifier,
                              TimeInterval interv,
                              QDateTime    last,
                              bool         active,
                              bool         readonly )
    : Dumpable( "BackupProfile", identifier ),
      mIdentifier( identifier ),
      mTimeToBackup( interv ),
      mLastBackupTime( last ),
      mbIsActivated( active ),
      mbReadOnly( readonly )
{
    kdDebug(0) << "BackupProfile::BackupProfile" << endl;
    mpBackup = backup;
    mpTimer = new QTimer();
    QObject::connect(
        mpBackup, SIGNAL(finishedBackup(bool)),
        this,     SLOT(slotFinishedBackup(bool))
        );
}

BackupProfile::~BackupProfile()
{
    kdDebug(0) << "BackupProfile::~BackupProfile" << endl;
    Q_ASSERT( mpBackup );
    Q_ASSERT( mpTimer );

    delete mpBackup;
    delete mpTimer;
}

void BackupProfile::activate()
{
    kdDebug(0) << "BackupProfile::activate" << endl;
    if ( mbReadOnly )
        return;
    if ( mbIsActivated )
        setupTimer();
}

void BackupProfile::restore()
{
    kdDebug(0) << "BackupProfile::restore" << endl;
    QFileInfo q( mpBackup->getArchiveUrl().path() );
    KURL      toRestore;

    if ( ! restoreActionIsSave() )
        return;
    toRestore = getToRestoreBackup();

    KArchive *archive = new KTar( toRestore.path() );
    if ( ! archive->open( IO_ReadOnly ) )
    {
        KMessageBox::error( 0L, i18n( "Sorry, somethings wrong with your archive."
                                      "I cannot restore the backup" ) );
        return;
    }
    KArchiveEntry *root = const_cast<KArchiveDirectory*>( archive->directory() );
    this->extractArchiveEntry( root, QString::null );
    archive->close();
}

QString BackupProfile::dumpInternal()
{
    kdDebug(0) << "BackupProfile::dumpInternal" << endl;
    QString tmp;
    tmp += QString( "mpTimer" ) + QString().setNum( (long) this->mpTimer ) + "\n";
    tmp += QString( "Identifier:  " ) + this->mIdentifier + "\n";
    tmp += this->mTimeToBackup.dump();
    tmp += QString( "LastBackup:  " ) + this->mLastBackupTime.toString(
        "yyyyMMddhhmmss" ) + "\n";

    if ( mbIsActivated ) tmp += QString( "mbIsActivated:     true\n" );
    else                 tmp += QString( "mbIsActivated:     false\n" );
    if ( mbReadOnly )     tmp += QString( "Readonly:          true\n" );
    else                 tmp += QString( "Readonly:          false\n" );

    tmp += mpBackup->dump();
    return tmp;
}

BackupProfile * BackupProfile::newBackupProfile(
    KURL      sourceUrl,
    KURL      backupUrl,
    QString   identifier,
    int       time,
    QString   measurement,
    bool      activ,
    QDateTime last
    )
{
    TimeInterval  t( (unsigned int) time, measurement );
    Backup *backup = new Backup(
        sourceUrl, backupUrl
        );
    return new BackupProfile( backup, identifier, t, last, activ );
}

BackupProfile *BackupProfile::newBackupProfileReadOnly(
    KURL      sourceUrl,
    KURL      backupUrl,
    QString   identifier,
    int       time,
    QString   measurement,
    bool      activ,
    QDateTime last
    )
{
    kdDebug(0) << "BackupProfile::newBackupProfileReadOnly" << " sourceUrl=" << sourceUrl.url() << " backupUrl=" << backupUrl.url() << " identifier=" << identifier << " time=" << time << " measurement=" << measurement << " activ=" << (activ ? "true" : "false" ) << endl;
    TimeInterval  t( (unsigned int) time, measurement );

    Backup *backup = new Backup(
        sourceUrl, backupUrl, true
        );
    return new BackupProfile(
        backup, identifier, t, last, activ, true
        );
}

BackupProfile *BackupProfile::newBackupProfileReadOnly(
    const BackupProfile * const b
    )
{
    kdDebug(0) << "BackupProfile::newBackupProfileReadOnly" << endl;
    TimeInterval  t( b->getTimeBetweenBackups() );
    Backup  *backup     = new Backup(
        b->getSourceUrl(), b->getArchiveUrl(), true
        );
    return new BackupProfile(
        backup, b->getIdentifier(), t, b->getLastTime(), b->isActivated(), true
        );
}

void BackupProfile::slotDoBackup()
{
    kdDebug(0) << "BackupProfile::slotDoBackup" << endl;
    slotDoOneTimeBackup();

    if ( mbIsActivated )  // Important: this one after mLastBackupTime = ...
        setupTimer();
}

void BackupProfile::slotDoOneTimeBackup()
{
    kdDebug(0) << "BackupProfile::slotDoOneTimeBackup" << endl;
    if ( mbReadOnly )
        return;

    doBackupIfNeeded();
}

void BackupProfile::slotFinishedBackup( bool successful )
{
    kdDebug(0) << "BackupProfile::slotFinishedBackup" << " successful=" << (successful ? "true" : "false" ) << endl;
    if ( successful )
        mLastBackupTime = QDateTime::currentDateTime();
    else
        KMessageBox::error(
            0L, i18n("The backup %1 was not made successfully.\n"
                     "Konserve will try again later").arg(mIdentifier)
            );
}

void BackupProfile::slotSearchForModifiedFiles( KIO::Job* /* aJob */,
                                                const KIO::UDSEntryList& list )
{
    kdDebug(0) << "BackupProfile::slotSearchForModifiedFiles" << endl;
    QDateTime changedTime;
    QDateTime lastBackupTime;
    // We use lastBackupTime instead of mLastBackupTime because
    // the last backup might have failed but mLastBackupTime was
    // increased. FIXME.

    KURL lastArchiveUrl = getToRestoreBackup();

    if (lastArchiveUrl.isEmpty())
    {
        mbFilesModified = true;
        return;
    }
    KFileItem lastBackup(
        KFileItem::Unknown, KFileItem::Unknown, lastArchiveUrl
        );
    lastBackupTime.setTime_t( lastBackup.time( KIO::UDS_MODIFICATION_TIME ) );

    KIO::UDSEntryListConstIterator it = list.begin();
    for (; it != list.end(); ++it)
    {
        KFileItem f( *it, mpBackup->getSourceUrl() );

        // Only filter out "..". If "." has been modified, a backup is needed
        if ( f.text() == ".." )
            continue;
        else
        {
            changedTime.setTime_t( f.time( KIO::UDS_MODIFICATION_TIME ) );

            // Backup needed, if even one file modified after last backup
            if( changedTime > lastBackupTime )
                mbFilesModified = true;
        }
    }
}

void BackupProfile::slotFinishedSearchingModifiedFiles( KIO::Job* aJob )
{
    kdDebug(0) << "BackupProfile::slotFinishedSearchingModifiedFiles" << " aJob=" << aJob << endl;
    if ( 0L != aJob )
    {
        if ( aJob->error() )
        {
            aJob->showErrorDialog();
            return;
        }
    }

    if( mbFilesModified )
    {
        mpBackup->doBackup();
    }
}

BackupProfile::BackupProfile( const BackupProfile& b )
    : QObject(),
      Dumpable( "BackupProfile", b.mIdentifier + " copy" ),
      mIdentifier( b.mIdentifier ),
      mTimeToBackup( b.mTimeToBackup ),
      mLastBackupTime( b.mLastBackupTime ),
      mbIsActivated( false ),
      mbReadOnly( true )
{
    kdDebug(0) << "BackupProfile::BackupProfile" << endl;
    mpTimer = new QTimer();
}

void BackupProfile::operator=( const BackupProfile& )
{
    kdDebug(0) << "BackupProfile::operator=" << endl;
    Q_ASSERT( false );
}

void BackupProfile::setupTimer()
{
    kdDebug(0) << "BackupProfile::setupTimer" << endl;
    int msecs = 0;

    if ( mpTimer->isActive() )
        mpTimer->stop();

    if ( !mLastBackupTime.isValid() )
        msecs = 10;
    else
        msecs = 1000 * mTimeToBackup.secs();

    mpTimer->singleShot( msecs, this, SLOT( slotDoBackup()));
}

void BackupProfile::extractArchiveEntry( KArchiveEntry* entry, QString path )
{
    kdDebug(0) << "BackupProfile::extractArchiveEntry" << " entry=" << entry << " path=" << path << endl;
    if ( entry->isDirectory() )
    {
        QDir dir;
        if ( dir.exists( path ) || QString::null == path )
            /* NOOP */;
        else
        {
            if ( ! dir.mkdir( path ) )
                KMessageBox::error(
                    0L, i18n( "Shit happend: Could not create dir: %1" ).arg( path )
                    );
        }

        QStringList list = ( dynamic_cast<KArchiveDirectory*>( entry ) )->entries();
        QStringList::iterator it = list.begin();
        while ( list.end() != it )
        {
            extractArchiveEntry(
                ( dynamic_cast<KArchiveDirectory*>( entry ) )->entry( *it ),
                path + "/" + *it
                );
            it++;
        }
    }
    else if ( entry->isFile() )
    {
        QFile file( path );
        if ( !file.open( IO_WriteOnly ) )
            KMessageBox::error(
                0L, i18n( "Shit happend! Could not open file: %1" ).arg( path )
                );
        else
        {
            file.writeBlock( ( dynamic_cast<KArchiveFile*>( entry ) )->data() );
            file.close();
        }
    }
    else
        KMessageBox::error(
            0L, i18n( "Shit happend! Entry not dir nor file!" )
            );
}

KURL BackupProfile::getToRestoreBackup()
{
    kdDebug(0) << "BackupProfile::getToRestoreBackup" << endl;
    QFileInfo qfinfo( mpBackup->getArchiveUrl().path() );
    QStringList list;
    KURL result;

    if ( qfinfo.isDir() )
    {
        list = qfinfo.dir().entryList(
            mpBackup->getSourceUrl().fileName() + "-??????????????.tar.gz",
            QDir::Files,
            QDir::Name | QDir::Reversed
            );
        return KURL( mpBackup->getArchiveUrl().url() + *list.begin() );
    }
    else
    {
        return mpBackup->getArchiveUrl();
    }
}

bool BackupProfile::restoreActionIsSave()
{
    kdDebug(0) << "BackupProfile::restoreActionIsSave" << endl;
    QFileInfo q( mpBackup->getSourceUrl().path() );

    if ( q.exists() )
    {
        KMessageBox::sorry(
            0L, i18n( "The url %1 to be restored still exists.\n"
                      "Please remove or rename ist first.").arg(
                          mpBackup->getSourceUrl().path()),
            i18n( "Url exists" )
            );
        return false;
    }
    return true;
}

void BackupProfile::doBackupIfNeeded()
{
    kdDebug(0) << "BackupProfile::doBackupIfNeeded" << endl;
    mbFilesModified = false;

    if ( QFileInfo( mpBackup->getSourceUrl().path() ).isFile() ||
         QFileInfo( mpBackup->getSourceUrl().path() ).isSymLink() )
    {
        QDateTime changedTime;
        KFileItem f(
            KFileItem::Unknown, KFileItem::Unknown, mpBackup->getSourceUrl()
            );
        changedTime.setTime_t( f.time( KIO::UDS_MODIFICATION_TIME ) );

        if( changedTime > mLastBackupTime )
            mbFilesModified = true;
        slotFinishedSearchingModifiedFiles(0L);
    }
    else if ( QFileInfo( mpBackup->getSourceUrl().path() ).isDir() )
    {
        KIO::Job* job = KIO::listRecursive( mpBackup->getSourceUrl(), false, true );
        connect(
            job, SIGNAL( entries( KIO::Job*, const KIO::UDSEntryList&)),
            SLOT( slotSearchForModifiedFiles( KIO::Job*, const KIO::UDSEntryList&))
            );
        connect(
            job, SIGNAL( result( KIO::Job* ) ),
            SLOT( slotFinishedSearchingModifiedFiles( KIO::Job* ) )
            );
    }
    else
    {
        Q_ASSERT( false );
    }
}

